from __future__ import print_function, division

import math
import numpy as np
import shapely.geometry
import shapely.affinity
import shapely.validation

from gdshelpers.parts import Port
from gdshelpers.parts.coupler_references import REFERENCES as COUPLER_REFERENCES
from gdshelpers.parts.text import Text
from gdshelpers.helpers import StandardLayers


class GratingCoupler(object):
    """
    A standard style radial grating coupler.

    This coupler starts with the ``width`` of the input/output waveguide. The input is widened to the opening
    angle and ended with a circle which radius is given by the first value in the list of grating lines. The taper
    radius is either measured absolutely starting at the part origin or relative to the minimum radius. The minimum
    radius is the radius needed to achieve the opening angle.

    All grating lines are given in a list which is the distance of the grating line to the last grating line. A list
    like ``(10, 2, 5, 3, 6)`` generates a coupler:

        * with taper radius 10
        * followed by a gap of 2 um and 5 um of material
        * followed by a gap of 4 um and 5 um of material

    In most cases it is too much work to calculate these grating intervals. The class functions
    :func:`make_traditional_coupler`, :func:`make_traditional_coupler_at_port` and
    :func:`make_traditional_coupler_from_database` will help you to create traditional couplers efficiently.

    :param origin: 2-element list or tuple specifying the ``(x, y)`` coordinates of the coupler.
    :type origin: list, tuple
    :param width: With of the waveguide
    :type width: float
    :param full_opening_angle: Full opening angle of the coupler in radiant.
    :type full_opening_angle: float
    :param grating_lines: List specifying the taper radius and the position of the grating edges.
    :param n_points: Number of points used per grating edge.
    :param extra_triangle_layer: If True, the taper triangle is painted separately and can be accessed via
        :func:`get_shapely_object_triangle`. Disabled by default.
    :type extra_triangle_layer: bool
    :param angle: Angle of the coupler in radiant. Defaults to -pi/2, hence the waveguide of the coupler points upwards.
    :param start_radius_absolute: If set to True, the first element of ``grating_lines`` is the absolute radius of
        the taper measured to its origin. An assertion error is thrown, if the radius is to small to achieve the
        full opening angle.
    """

    def __init__(self, origin, angle, width, full_opening_angle, grating_lines, n_points=300,
                 extra_triangle_layer=False, start_radius_absolute=False):
        assert len(grating_lines) % 2 == 1, 'An odd number of grating lines is needed.'
        assert len(grating_lines) > 0, 'At least one element is needed in grating_lines'

        self._origin = origin
        self._angle = angle
        self._width = width
        self._opening_angle = full_opening_angle / 2
        self._grating_lines = grating_lines
        self._points = n_points

        self._triangle_polygon = None
        self._grating_polygon = None
        self._extra_triangle_layer = extra_triangle_layer
        self._start_radius_absolute = start_radius_absolute
        self._maximal_radius = None

        self._traditional_parameters = None

    @classmethod
    def make_traditional_coupler(cls, origin, width, full_opening_angle, grating_period, grating_ff, n_gratings,
                                 ap_max_ff=1.0, n_ap_gratings=0, taper_length=None, extra_triangle_layer=False,
                                 angle=-np.pi / 2, n_points=394, implement_cadence_ff_bug=True):
        """
        Generate a traditional coupler as the ``lib/coupler/`` functions did. But this version is versatile enough
        to generate all kinds of couplers, including:

            - constant fill factor couplers
            - apodized couplers (where the fill factor is varied from ap_max_ff to grating_ff over
                                 n_ap_gratings gratings.)

        All couplers can have the taper triangle separated onto its own layer for better e-beam dose.

        :param origin: 2-element list or tuple specifying the ``(x, y)`` coordinates of the coupler.
        :type origin: list, tuple
        :param width: With of the waveguide
        :type width: float
        :param full_opening_angle: Full opening angle of the coupler in radiant.
        :type full_opening_angle: float
        :param grating_period: The grating period of the coupler.
        :type grating_period: float
        :param grating_ff: The fill factor of the coupler between 0 and 1.
        :type grating_ff: float
        :param n_gratings: Number of gratings.
        :type n_gratings: int
        :param ap_max_ff: Maximum fill factor of apodized gratings.
        :param n_ap_gratings: Number of apodized gratings.
        :type n_ap_gratings: int
        :param taper_length: If not None, this sets the radius if the inner taper. Otherwise the length will
                             be equal to the grating length.
        :param n_points: Number of points used per grating edge.
        :param extra_triangle_layer: If True, the taper triangle is painted seperatly and can be accessed via
            :func:`get_shapely_object_triangle`. Disabled by default.
        :type extra_triangle_layer: bool
        :param angle: Angle of the coupler in radiant. Defaults to -pi/2, hence the waveguide of the coupler points
                      upwards.
        :parameter implement_cadence_ff_bug:
        :type implement_cadence_ff_bug: bool

        :return: The constructed traditional grating coupler.
        :rtype: GratingCoupler
        """

        assert 0 < grating_ff < 1 and 0 < ap_max_ff <= 1, 'Fill factors must be between 0 and 1.'

        # We first generate the list for the grating radii.
        #
        radii = list()

        # Add apodized gratings
        if not implement_cadence_ff_bug:
            apodized_ffs = np.linspace(ap_max_ff, grating_ff, n_ap_gratings)
        else:
            apodized_ffs = np.linspace(ap_max_ff, grating_ff, n_ap_gratings + 1)[1:]

        for ap_ff in apodized_ffs:
            radii.append(grating_period * (1 - ap_ff))
            radii.append(grating_period * ap_ff)

        # Add simple gratings with constant ff
        radii.extend([grating_period * (1 - grating_ff), grating_period * grating_ff] * n_gratings)

        # Insert the inner radius
        radii.insert(0, taper_length if taper_length is not None else sum(radii))

        obj = cls(origin, angle, width, full_opening_angle, radii, n_points // 2,
                  start_radius_absolute=True, extra_triangle_layer=extra_triangle_layer)
        obj._traditional_parameters = {
            'grating_period': grating_period,
            'grating_ff': grating_ff,
            'n_gratings': n_gratings,
            'ap_max_ff': ap_max_ff,
            'ap_max_ff_true': apodized_ffs[0] if n_ap_gratings else None,
            'n_ap_gratings': n_ap_gratings,
            'taper_length': taper_length,
            'implement_cadence_ff_bug': implement_cadence_ff_bug,
            'full_opening_angle': full_opening_angle
        }

        return obj

    @classmethod
    def make_traditional_coupler_at_port(cls, port, **kwargs):
        """
        Make a traditional coupler at a port.

        This function is identical to :func:`make_traditional_coupler`. Parameters of the port
        can also be overwritten via keyword arguments.

        :param port: The port at which the coupler shall be created.
        :type port: Port
        :param kwargs: Keyword arguments passed to :func:`make_traditional_coupler`.
        :return: The constructed traditional grating coupler.
        :rtype: GratingCoupler
        """

        if 'width' not in kwargs:
            kwargs['width'] = port.width

        if 'origin' not in kwargs:
            kwargs['origin'] = port.origin

        if 'angle' not in kwargs:
            kwargs['angle'] = port.angle

        return cls.make_traditional_coupler(**kwargs)

    @classmethod
    def make_traditional_coupler_from_database(cls, origin, width, db_id, wavelength, **kwargs):
        # Get the parameters
        parameters = dict(COUPLER_REFERENCES[db_id]['parameters'])

        # Update the wavelength with calculated wavelength
        parameters['grating_period'] = parameters['grating_period'](wavelength)

        # Update the parameters with user specified parameters
        parameters.update(kwargs)

        return cls.make_traditional_coupler(origin, width, **parameters)

    @classmethod
    def make_traditional_coupler_from_database_at_port(cls, port, db_id, wavelength, **kwargs):
        return cls.make_traditional_coupler_from_database(port.origin, port.width, db_id, wavelength, angle=port.angle,
                                                          **kwargs)

    @property
    def maximal_radius(self):
        if not self._maximal_radius:
            self._generate()
        return self._maximal_radius

    def _generate(self):
        # For D1 continuous change from port waveguide side to the opening "triangle" we use circles.
        # We first generate this triangle and add the grating arcs later
        alpha = math.pi / 2 - self._opening_angle
        c_radius = -math.sin(alpha) * self._width / 2 / (math.sin(alpha) - 1)

        tmp_a, tmp_b = c_radius, (c_radius + self._width / 2)
        minimum_taper_radius = math.sqrt(tmp_a ** 2 + tmp_b ** 2 - 2 * tmp_a * tmp_b * math.cos(self._opening_angle))

        radius = self._grating_lines[0] if self._start_radius_absolute else minimum_taper_radius + self._grating_lines[
            0]
        assert (not self._start_radius_absolute or
                self._grating_lines[0] >= minimum_taper_radius), 'Start radius is smaller than minimum start radius!'

        # noinspection PyTypeChecker
        phi = np.linspace(0, self._opening_angle, 90) - math.pi / 2
        upper_half = [np.cos(phi) * c_radius, np.sin(phi) * c_radius + c_radius + self._width / 2]
        # noinspection PyTypeChecker
        phi = np.linspace(-self._opening_angle, 0, 90) + math.pi / 2
        lower_half = [np.cos(phi) * c_radius, np.sin(phi) * c_radius - c_radius - self._width / 2]

        phi = np.linspace(self._opening_angle, -self._opening_angle, self._points)
        inner_circle = [np.cos(phi) * radius, np.sin(phi) * radius]

        points = np.hstack((upper_half, inner_circle, lower_half)).T
        opening_polygon = shapely.geometry.Polygon(points)

        grating_points = []

        phi = np.linspace(-self._opening_angle + self._angle, self._opening_angle + self._angle, self._points)
        for addition_r in self._grating_lines[1:]:
            radius += addition_r
            grating_points.append([np.cos(phi) * radius + self._origin[0], np.sin(phi) * radius + self._origin[1]])
        self._maximal_radius = radius

        # Separate into pairs and create polygon
        grating_polygons = []
        for inner_half, outer_half in zip(grating_points[::2], grating_points[1::2]):
            inner_half = np.array(inner_half)
            if inner_half.shape[0] > 1:
                inner_half = inner_half.T
            outer_half = np.array(outer_half).T

            # Add both paths together to form the polygon. We also need to reverse the order of one side, so that
            # the polygon is not illegally twisted.
            points = np.vstack((inner_half, outer_half[::-1, :]))

            grating_polygons.append((points, []))

        # Merge all polygons to one big shapely object
        triangle_polygon = opening_polygon
        triangle_polygon = shapely.affinity.rotate(triangle_polygon, self._angle, origin=[0, 0], use_radians=True)
        triangle_polygon = shapely.affinity.translate(triangle_polygon, self._origin[0], self._origin[1])
        self._triangle_polygon = triangle_polygon

        grating_polygon = shapely.geometry.MultiPolygon(grating_polygons)
        self._grating_polygon = grating_polygon

    @property
    def origin(self):
        return self.port.origin

    @property
    def width(self):
        return self.port.width

    @property
    def port(self):
        """
        The port of the coupler.

        :return: The coupler port.
        :rtype: Port
        """
        return Port(self._origin, self._angle, self._width).inverted_direction

    def get_shapely_object(self):
        """
        Get the Shapely object of this coupler.

        If requested during creation of the coupler, the tapered triangle is not included in this Shapely
        object. It has to be acquired via :func:`get_shapely_object_triangle`.

        :return: A Shapely object.
        """
        if self._grating_polygon is None or self._triangle_polygon is None:
            self._generate()

        if self._extra_triangle_layer:
            return self._grating_polygon
        else:
            return self._triangle_polygon.union(self._grating_polygon)

    def get_shapely_object_triangle(self):
        """
        Get the Shapely object of this coupler, but only the tapered triangle.

        :return: A Shapely object.
        """
        if self._triangle_polygon is None:
            self._generate()
        return self._triangle_polygon

    def get_description_str(self, grating_period=True, fill_factor=True, taper_length=True, opening_angle=True):
        desc = list()

        if self._traditional_parameters:
            if grating_period:
                desc.append(('gp', '%.2f' % self._traditional_parameters['grating_period']))

            if fill_factor:
                desc.append(('ff', '%.2f' % self._traditional_parameters['grating_ff']))

            desc.append(('n', '%i' % self._traditional_parameters['n_gratings']))

            if self._traditional_parameters['n_ap_gratings']:
                desc.append(('n_ap', '%i' % self._traditional_parameters['n_ap_gratings']))

                if fill_factor:
                    desc.append(('ff_ap', '%.2f' % self._traditional_parameters['ap_max_ff']))

                    if self._traditional_parameters['implement_cadence_ff_bug']:
                        desc.append(('ff_ap_true', '%.2f' % self._traditional_parameters['ap_max_ff_true']))

            if taper_length and self._traditional_parameters['taper_length']:
                desc.append(('tl', self._traditional_parameters['taper_length']))

            if opening_angle:
                desc.append(('angle', '%.2f' % np.rad2deg(self._traditional_parameters['full_opening_angle'])))

        else:
            raise NotImplementedError('Description not yet implemented for non-traditional couplers')

        return '\n'.join(['%s:%s' % x for x in desc])

    def get_description_text(self, height=3., space=10, side='right', **desc_options):
        assert side in ['left', 'right'], 'side parameter must be left or right'

        offset = abs(np.cos(np.pi / 2 - self._opening_angle)) * self.maximal_radius + space
        if side == 'right':
            offset *= -1.
            alignment = 'left-top'
        if side == 'left':
            alignment = 'right-top'

        text_port = self.port.parallel_offset(offset).rotated(3 / 2 * np.pi)
        text = Text(text_port.origin, height, text=self.get_description_str(**desc_options),
                    alignment=alignment, angle=text_port.angle)
        return text


def _example():
    from gdshelpers.geometry.chip import Cell
    from gdshelpers.geometry import geometric_union

    # Generate a coupler, which ought to be identical to
    # coupler_sn330_1550_bf_ff_ap( 0:0 1550 0.7 22 0.96 10 22 1 "active")
    # also known as
    # coupler_bf_ap_mff( 0:0 1 40.0 1.13 0.7 22 0.96 10 22 200 "active")
    coupler = GratingCoupler.make_traditional_coupler([150, 0], 1, np.deg2rad(40), 1.13, 0.7,
                                                      22, 0.96, 10, 22, angle=-np.pi)

    coupler2 = GratingCoupler.make_traditional_coupler_from_database([0, 0], 1, 'sn330', 1550)

    print(coupler.get_description_str())

    whole_layout = (coupler, coupler2)

    layout = Cell('LIBRARY')
    cell = Cell('TOP')
    cell.add_to_layer(1, *whole_layout)
    cell.add_to_layer(StandardLayers.parnamelayer1, coupler.get_description_text())
    cell.add_to_layer(StandardLayers.parnamelayer1, coupler2.get_description_text())

    cell.add_to_layer(1, coupler.get_description_text())
    cell.add_to_layer(1, coupler2.get_description_text())
    cell.add_to_layer(1, coupler2.get_description_text(side='left'))
    layout.add_cell(cell)

    # We could also easily generate a dark field layout out of this
    def make_dark_field(obj, buffer_size=1.5):
        return obj.buffer(buffer_size).difference(obj)

    cell_df = Cell('TOP_DF')
    cell_df.add_to_layer(2, make_dark_field(geometric_union(whole_layout)))
    layout.add_cell(cell_df)

    layout.save('coupler.gds')
    cell.show()


if __name__ == '__main__':
    _example()



import unittest
import math

from gdshelpers.parts.waveguide import Waveguide


class WaveguideTestCase(unittest.TestCase):
    def test_waveguide_construction(self):
        wg = Waveguide([0, 0], 0, 1)
        self.assertAlmostEqual(wg.length, 0)

        # Add straight waveguide
        wg.add_straight_segment(10)
        self.assertAlmostEqual(wg.x, 10)
        self.assertAlmostEqual(wg.y, 0)
        self.assertAlmostEqual(wg.angle, 0)
        self.assertAlmostEqual(wg.length, 10)

        # Add upwards 90 degree bend
        wg.add_bend(math.pi / 2, 5)
        self.assertAlmostEqual(wg.current_port.origin[0], 15)
        self.assertAlmostEqual(wg.current_port.origin[1], 5)
        self.assertAlmostEqual(wg.current_port.angle, math.pi / 2)
        self.assertAlmostEqual(wg.length_last_segment, math.pi / 2 * 5, delta=10e-5)

        # Add another arc, so that we are finally pointing 180 degree backwards
        wg.add_arc(math.pi, 5)
        self.assertAlmostEqual(wg.current_port.origin[0], 10)
        self.assertAlmostEqual(wg.current_port.origin[1], 10)
        self.assertAlmostEqual(abs(wg.current_port.angle), math.pi)



"""
For documentation of GDSII see: http://www.buchanan1.net/stream_description.html
"""

import math
import datetime
from struct import pack
from io import BytesIO
import numpy as np

from shapely.geometry import Polygon


def _real_to_8byte(value):
    if value == 0:
        return b'\x00' * 8
    exponent = int((math.log(abs(value), 16) + 1) // 1)
    mantissa = int(abs(value) * 16. ** (14 - exponent))
    return ((((0b1 if value < 0 else 0b0) + exponent + 64) << 56) + mantissa).to_bytes(8, 'big')


def _cell_to_gdsii_binary(cell, grid_steps_per_unit, max_points, max_line_points, timestamp):
    with BytesIO() as b:
        b.write(pack('>14H', 28, 0x0502, *timestamp.timetuple()[:6] * 2))
        # BGNSTR INTEGER_2 time_modification time_last_access
        name = cell.name + '\0' * (len(cell.name) % 2)
        b.write(pack('>2H', 4 + len(name), 0x0606) + name.encode('ascii'))  # STRNAME STRING cell_name

        for layer, polygons in cell.get_fractured_layer_dict(max_points, max_line_points).items():
            for polygon in polygons:
                if not isinstance(polygon, Polygon):
                    import warnings
                    warnings.warn(
                        'Shapely object of type ' + str(type(polygon)) + ' not convertible to GDSII, skipping...')
                    continue
                if polygon.interiors:
                    raise AssertionError('GDSII only supports polygons without holes')
                xy = np.round(
                    np.array(
                        list(polygon.exterior.coords) + [polygon.exterior.coords[0]]) * grid_steps_per_unit).astype(
                    '>i4')
                b.write(pack('>8H', 4, 0x0800,  # BOUNDARY NO_DATA
                             6, 0x0D02, layer,  # LAYER INTEGER_2 layer
                             6, 0x0E02, layer))  # DATATYPE INTEGER_2 datatype
                for start in range(0, xy.shape[0], 8191):  # Split in Blocks of 8191 points
                    stop = min(start + 8191, xy.shape[0])
                    b.write(pack('>2H', 4 + 8 * (stop - start), 0x1003))  # XY INTEGER_4
                    b.write(xy[start:stop].tobytes())  # coords of polygon
                b.write(pack('>2H', 4, 0x1100))  # ENDEL NO_DATA

        for ref in cell.cells:
            aref = not (ref['columns'] == 1 and ref['rows'] == 1 and not ref['spacing'])
            name = ref['cell'].name + '\0' if len(ref['cell'].name) % 2 != 0 else ref['cell'].name
            if aref:
                b.write(pack('>2H', 4, 0x0B00))  # AREF NO_DATA
            else:
                b.write(pack('>2H', 4, 0x0A00))  # SREF NO_DATA
            b.write(pack('>2H', 4 + len(name), 0x1206) + name.encode('ascii'))  # SNAME STRING ref_cell_name
            if (ref['angle'] is not None) or (ref['magnification'] is not None) or ref['x_reflection']:
                b.write(pack('>3H', 6, 0x1A01, 1 << 15 if ref['x_reflection'] else 0))  # STRANS BIT_ARRAY bit15=1
                if ref['magnification'] is not None:
                    b.write(pack('>2H', 12, 0x1B05) + _real_to_8byte(ref['magnification']))  # MAG REAL_8
                if ref['angle'] is not None:
                    b.write(pack('>2H', 12, 0x1C05) + _real_to_8byte(np.rad2deg(ref['angle'])))  # ANGLE REAL_8
            if aref:
                b.write(pack('>2H2h', 8, 0x1302, ref['columns'], ref['rows']))  # COLROW INTEGER_2 spacing
            b.write(pack('>2H', 28 if aref else 12, 0x1003) + np.round(
                np.array(ref['origin']) * grid_steps_per_unit).astype(
                '>i4').tobytes())  # XY INTEGER_8 origin
            if aref:
                b.write((np.round(np.array((ref['spacing'][0] * ref['columns'], 0)) +
                                  np.array(ref['origin'])) * grid_steps_per_unit).astype(
                    '>i4').tobytes())  # XY INTEGER_8 edge_x
                b.write((np.round(np.array((0, ref['spacing'][1] * ref['rows'])) +
                                  np.array(ref['origin'])) * grid_steps_per_unit).astype(
                    '>i4').tobytes())  # XY INTEGER_8 edge_y

            b.write(pack('>2H', 4, 0x1100))  # ENDEL NO_DATA

        b.write(pack('>2H', 4, 0x0700))  # ENDEST NO_DATA

        return b.getvalue()


def write_cell_to_gdsii_file(outfile, cell, unit=1e-6, grid_steps_per_unit=1000, max_points=4000, max_line_points=4000,
                             timestamp=None, parallel=False):
    name = 'gdshelpers_exported_library'
    grid_step_unit = unit / grid_steps_per_unit
    timestamp = datetime.datetime.now() if timestamp is None else timestamp

    cells = []
    cell_names = []

    def add_cells_to_unique_list(start_cell):
        cells.append(start_cell)
        cell_names.append(start_cell.name)
        for c in start_cell.cells:
            if c['cell'] not in cells:
                if c['cell'].name in cell_names:
                    raise AssertionError(
                        'Each cell name must be unique, "{}" is used more than once'.format(c['cell'].name))
                add_cells_to_unique_list(c['cell'])

    add_cells_to_unique_list(cell)

    name = name + '\0' * (len(name) % 2)  # Strings always have even length
    outfile.write(pack('>3H', 6, 0x0002, 0x258))  # HEADER INTEGER_2 v6.0
    outfile.write(pack('>14H', 28, 0x0102, *timestamp.timetuple()[:6] * 2))
    # BGNLIB INTEGER_2 time_modification time_last_access
    outfile.write(pack('>2H', 4 + len(name), 0x0206) + name.encode('ascii'))  # LIBNAME STRING libname
    outfile.write(pack('>2H', 20, 0x0305) + _real_to_8byte(grid_step_unit / unit) + _real_to_8byte(grid_step_unit))
    # UNITS REAL_8 1/grid_steps_per_unit grid_step_unit
    if parallel:
        from concurrent.futures import ProcessPoolExecutor
        with ProcessPoolExecutor() as pool:
            num = len(cells)
            for binary in pool.map(_cell_to_gdsii_binary, cells, (grid_steps_per_unit,) * num, (max_points,) * num,
                                   (max_line_points,) * num, (timestamp,) * num):
                outfile.write(binary)
    else:
        for c in cells:
            outfile.write(_cell_to_gdsii_binary(c, grid_steps_per_unit, max_points, max_line_points, timestamp))
    outfile.write(pack('>2H', 4, 0x0400))  # ENDLIB N0_DATA


if __name__ == '__main__':
    from gdshelpers.parts.port import Port
    from gdshelpers.parts.waveguide import Waveguide
    from gdshelpers.geometry.chip import Cell

    device_cell = Cell('cell')
    start_port = Port(origin=(10, 0), width=1, angle=0)
    waveguide = Waveguide.make_at_port(start_port)
    for i_bend in range(9):
        waveguide.add_bend(angle=np.pi, radius=60 + i_bend * 40)
    device_cell.add_dlw_taper_at_port('A', 2, waveguide.in_port, 30)
    device_cell.add_dlw_taper_at_port('B', 2, waveguide.current_port, 30)
    device_cell.add_to_layer(1, waveguide)

    sub_cell = Cell('sub_cell')
    sub_cell.add_to_layer(1, waveguide)

    device_cell.add_cell(sub_cell, origin=(10, 10), angle=np.pi / 2)

    with open('gdsii_export.gds', 'wb') as file:
        write_cell_to_gdsii_file(file, device_cell, parallel=True)



import unittest

from gdshelpers.helpers.alignment import Alignment


class AlignmentTestCase(unittest.TestCase):
    def test_calculate_offset(self):
        alignment = Alignment('left-bottom')
        self.assertTupleEqual(tuple(alignment.calculate_offset(((1, 2), (11, 12)))), (-1, -2))
        alignment = Alignment('center-bottom')
        self.assertTupleEqual(tuple(alignment.calculate_offset(((1, 2), (11, 12)))), (-6, -2))
        alignment = Alignment('right-center')
        self.assertTupleEqual(tuple(alignment.calculate_offset(((1, 2), (11, 12)))), (-11, -7))
        alignment = Alignment('right-top')
        self.assertTupleEqual(tuple(alignment.calculate_offset(((1, 2), (11, 12)))), (-11, -12))


if __name__ == '__main__':
    unittest.main()



"""
Parts for Optical Field Writable Arrays (OFWA)
"""

import numpy as np

from gdshelpers.parts import Port
from gdshelpers.parts.waveguide import Waveguide
from gdshelpers.parts.marker import DLWMarker

from gdshelpers.geometry import geometric_union
from gdshelpers.helpers import normalize_phase


class MultiPortSwitch(object):
    def __init__(self, origin, angle, in_ports, out_ports, port_spacing, taper_length, taper_function, radius,
                 wg_bend_radius, displacement=0., minimal_final_spacing=None):
        self._origin = origin
        self._angle = angle
        self._in_ports = in_ports
        self._out_ports = out_ports
        self._port_spacing = port_spacing
        self._taper_length = taper_length
        self._taper_function = taper_function
        self._taper_options = {}
        self._radius = radius
        self._wg_bend_radius = wg_bend_radius
        self._displacement = displacement
        self._minimal_final_spacing = minimal_final_spacing

        self._in_wgs = None
        self._out_wgs = None

    @classmethod
    def make_at_in_port(cls, port, in_port_idx, **kwargs):
        tmp_mps = cls([0, 0], port.angle, **kwargs)
        return cls(-tmp_mps.get_in_port(in_port_idx).origin + port.origin, port.angle, **kwargs)

    @classmethod
    def make_at_out_port(cls, port, out_port_idx, **kwargs):
        tmp_mps = cls([0, 0], port.inverted_direction.angle, **kwargs)
        return cls(-tmp_mps.get_out_port(out_port_idx).origin + port.origin, port.inverted_direction.angle, **kwargs)

    @property
    def origin(self):
        return np.array(self._origin)

    @property
    def angle(self):
        return self._angle

    @property
    def _angular_spacing(self):
        return self._port_spacing / self._radius

    def _calculate(self, do_in_wgs=True, do_out_wgs=True):
        angular_spacing = self._angular_spacing
        assert angular_spacing * self._out_ports < np.pi, 'Not enough space for output ports'
        assert angular_spacing * self._in_ports < np.pi, 'Not enough space for input ports'

        if do_out_wgs:
            # Do the output side
            out_origin_port = Port(self._origin, self._angle, 1.).longitudinal_offset(-self._displacement / 2)
            out_fanout_wgs = [Waveguide.make_at_port(out_origin_port.rotated(angle).longitudinal_offset(self._radius))
                              for angle in (np.arange(self._out_ports) - (self._out_ports - 1) / 2.) * angular_spacing]

            for wg in out_fanout_wgs:
                wg.add_parameterized_path(path=lambda t: [t * self._taper_length, np.zeros_like(t)],
                                          path_derivative=lambda t: [np.ones_like(t) * self._taper_length,
                                                                     np.zeros_like(t)],
                                          path_function_supports_numpy=True,
                                          width=self._taper_function, **self._taper_options)

            if self._minimal_final_spacing is None:
                for wg in out_fanout_wgs:
                    wg.add_bend(normalize_phase(-wg.angle + self._angle), self._wg_bend_radius)

            else:
                offsets = (np.arange(self._out_ports) - (self._out_ports - 1) / 2.) * self._minimal_final_spacing
                final_port_heights = [out_origin_port.parallel_offset(offset) for offset in offsets]

                for wg, final_port_height, offset in zip(out_fanout_wgs, final_port_heights, offsets):
                    if np.isclose(offset, 0):
                        continue

                    try:
                        wg.add_route_single_circle_to_port(final_port_height.inverted_direction, on_line_only=True)
                    except AssertionError:
                        # No curve possible, use normal bend
                        wg.add_bend(normalize_phase(-wg.angle + self._angle), self._wg_bend_radius)

            final_ports = [wg.current_port for wg in out_fanout_wgs]
            for wg in out_fanout_wgs:
                wg.add_straight_segment_until_level_of_port(final_ports)
            self._out_wgs = out_fanout_wgs

        if do_in_wgs:
            # Do the input side
            in_origin_port = Port(self._origin, self._angle + np.pi, 1.).longitudinal_offset(-self._displacement / 2)
            in_fanout_wgs = [Waveguide.make_at_port(in_origin_port.rotated(angle).longitudinal_offset(self._radius))
                             for angle in (np.arange(self._in_ports) - (self._in_ports - 1) / 2.) * angular_spacing]

            for wg in in_fanout_wgs:
                wg.add_parameterized_path(path=lambda t: [t * self._taper_length, np.zeros_like(t)],
                                          path_derivative=lambda t: [np.ones_like(t) * self._taper_length,
                                                                     np.zeros_like(t)],
                                          path_function_supports_numpy=True,
                                          width=self._taper_function, **self._taper_options)

            if self._minimal_final_spacing is None:
                for wg in in_fanout_wgs:
                    wg.add_bend(normalize_phase(-wg.angle + self._angle - np.pi), self._wg_bend_radius)

            else:
                offsets = (np.arange(self._in_ports) - (self._in_ports - 1) / 2.) * self._minimal_final_spacing
                final_port_heights = [in_origin_port.parallel_offset(offset) for offset in offsets]

                for wg, final_port_height, offset in zip(in_fanout_wgs, final_port_heights, offsets):
                    if np.isclose(offset, 0):
                        continue

                    # wg.add_route_single_circle_to_port(final_port_height.inverted_direction, on_line_only=True)

                    try:
                        wg.add_route_single_circle_to_port(final_port_height.inverted_direction, on_line_only=True)
                    except AssertionError:
                        # No curve possible, use normal bend
                        wg.add_bend(normalize_phase(-wg.angle + self._angle - np.pi), self._wg_bend_radius)

            final_ports = [wg.current_port for wg in in_fanout_wgs]
            for wg in in_fanout_wgs:
                wg.add_straight_segment_until_level_of_port(final_ports)

            self._in_wgs = in_fanout_wgs

    def get_in_port(self, idx):
        if not self._in_wgs:
            self._calculate(do_in_wgs=True, do_out_wgs=False)

        assert 0 <= idx <= self._in_ports, 'Invalid input port number'
        return self._in_wgs[idx].current_port

    def get_out_port(self, idx):
        if not self._out_wgs:
            self._calculate(do_in_wgs=False, do_out_wgs=True)

        assert 0 <= idx <= self._out_ports, 'Invalid input port number'
        return self._out_wgs[idx].current_port

    def get_dlw_in_port(self, idx):
        if not self._in_wgs:
            self._calculate(do_in_wgs=True, do_out_wgs=False)

        assert 0 <= idx <= self._in_ports, 'Invalid input port number'
        return self._in_wgs[idx].in_port

    def get_dlw_out_port(self, idx):
        if not self._out_wgs:
            self._calculate(do_in_wgs=False, do_out_wgs=True)

        assert 0 <= idx <= self._out_ports, 'Invalid input port number'
        return self._out_wgs[idx].in_port

    @property
    def in_ports(self):
        return list((self.get_in_port(idx) for idx in range(self._in_ports)))

    @property
    def out_ports(self):
        return list((self.get_out_port(idx) for idx in range(self._out_ports)))

    @property
    def dlw_in_ports(self):
        return list((self.get_dlw_in_port(idx) for idx in range(self._in_ports)))

    @property
    def dlw_out_ports(self):
        return list((self.get_dlw_out_port(idx) for idx in range(self._out_ports)))

    @property
    def marker_positions(self):
        center_port = Port(self._origin, self._angle, 1.)

        angle = max((2 + max(self._in_ports, self._out_ports)) / 2. * self._angular_spacing, np.pi / 4)
        radius = max(self._radius, 40)
        return [center_port.rotated(angle).longitudinal_offset(radius).origin,
                center_port.rotated(angle).longitudinal_offset(-radius).origin,
                center_port.rotated(-angle).longitudinal_offset(radius).origin,
                center_port.rotated(-angle).longitudinal_offset(-radius).origin]

    def get_shapely_object(self):
        if not self._in_wgs or not self._out_wgs:
            self._calculate(do_in_wgs=True, do_out_wgs=True)

        markers = [DLWMarker(pos) for pos in self.marker_positions]

        return geometric_union(self._out_wgs + self._in_wgs + markers)


def _example():
    import gdsCAD.core
    from gdshelpers.geometry import convert_to_gdscad
    from gdshelpers.parts.waveguide import Waveguide

    mpsv2 = MultiPortSwitch(origin=[0, 0],
                            angle=np.deg2rad(0.),
                            in_ports=4, out_ports=7,
                            port_spacing=10,
                            taper_length=40, taper_function=lambda t: t * (1.0 - 0.05) + 0.05,
                            radius=50., displacement=0.,
                            wg_bend_radius=40., minimal_final_spacing=20)

    wg = Waveguide([100, 100], np.pi / 4, 1)
    wg.add_straight_segment(100)

    # mpsv2_2 = MultiPortSwitch.make_at_out_port(wg.current_port, 3,
    #                                            in_ports=2, out_ports=7,
    #                                            port_spacing=10,
    #                                            taper_length=10, taper_function=lambda t: t * (1.0 - 0.05) + 0.05,
    #                                            radius=50., displacement=00.,
    #                                            wg_bend_radius=40., )

    cell = gdsCAD.core.Cell('TEST')
    cell.add(convert_to_gdscad([mpsv2, wg, ]))
    cell.show()


if __name__ == '__main__':
    _example()



import filecmp
import unittest
import numpy as np
from shapely.affinity import translate, rotate

from gdshelpers.parts.waveguide import Waveguide
from gdshelpers.geometry.chip import Cell
from gdshelpers.parts.pattern_import import GDSIIImport


class GdsTestCase(unittest.TestCase):
    def test_export_import(self):
        waveguide = Waveguide([0, 0], 0, 1)
        for i_bend in range(9):
            waveguide.add_bend(angle=np.pi, radius=60 + i_bend * 40)
        offset = (10, 10)
        angle = np.pi

        cell = Cell('test')
        cell.add_to_layer(1, waveguide)

        sub_cell = Cell('sub_cell')
        sub_cell.add_to_layer(2, waveguide)
        cell.add_cell(sub_cell, origin=(0, 0), angle=0)

        sub_cell2 = Cell('sub_cell2')
        sub_cell2.add_to_layer(3, waveguide)
        cell.add_cell(sub_cell2, origin=offset, angle=angle)

        cell.save(library='gdshelpers', grid_steps_per_micron=10000)

        def assert_almost_equal_shapely(a, b, tolerance=2e-4):
            self.assertTrue(a.buffer(tolerance).contains(b))
            self.assertTrue(b.buffer(tolerance).contains(a))

        assert_almost_equal_shapely(
            waveguide.get_shapely_object(), GDSIIImport('test.gds', 'test', 1).get_shapely_object())

        assert_almost_equal_shapely(
            waveguide.get_shapely_object(), GDSIIImport('test.gds', 'test', 2).get_shapely_object())

        assert_almost_equal_shapely(
            translate(rotate(waveguide.get_shapely_object(), angle, use_radians=True, origin=(0, 0)),
                      *offset), GDSIIImport('test.gds', 'test', 3).get_shapely_object())

        self.assertTrue(GDSIIImport('test.gds', 'test', 1, 2).get_shapely_object().is_empty)

    def test_parallel_export(self):
        waveguide = Waveguide([0, 0], 0, 1)
        for i_bend in range(9):
            waveguide.add_bend(angle=np.pi, radius=60 + i_bend * 40)

        cells = [Cell('main')]
        for i in range(10):
            cell = Cell('sub_cell_' + str(i))
            cell.add_to_layer(waveguide)
            cells[-1].add_cell(cell, (10, 10))

        cells[0].save('serial.gds', library='gdshelpers', parallel=False)
        cells[0].save('parallel.gds', library='gdshelpers', parallel=True)

        self.assertTrue(filecmp.cmp('serial.gds', 'parallel.gds'))



# Generate qrcodes
# - Requires the python-qrcode module from https://github.com/lincolnloop/python-qrcode

try:
    # noinspection PyUnresolvedReferences
    import qrcode
    import qrcode.image.base
except ImportError:
    raise ImportError('python-qrcode package needed for QR code generation')

import shapely.geometry
import shapely.ops
import shapely.affinity
import numpy as np

from gdshelpers.helpers.alignment import Alignment


class ShapelyImageFactory(qrcode.image.base.BaseImage):
    """
    Output factory for qrcode, generating Shapely polygons.

    Probably you will not need to use it directly. The :class:`.QRCode` class provides a simple to use part
    based on this image factory.
    """
    kind = "shapely"

    # noinspection PyAttributeOutsideInit
    def new_image(self, origin=(0, 0), scale_factor=1.0e-3, **kwargs):
        img = list()
        self._idr = img
        self._origin = origin
        self._scale_factor = scale_factor
        return img

    @property
    def origin(self):
        return np.asarray(self._origin)

    @property
    def size(self):
        pixels = self.width + 2 * self.border
        return np.ones((2,)) * pixels * self._scale_factor

    def drawrect(self, row, col):
        origin = self._origin

        box_coords = np.array(self.pixel_box(row, col)) * self._scale_factor
        box = shapely.geometry.box(box_coords[0][0] + origin[0],
                                   self.size[1] - box_coords[0][1] + origin[1],
                                   box_coords[1][0] + self._scale_factor + origin[0],
                                   self.size[1] - box_coords[1][1] - self._scale_factor + origin[1])
        self._idr.append(box)

    def get_shapely_object(self):
        return shapely.ops.cascaded_union(self._idr)


class QRCode(object):
    """
    Quick Response (QR) code part.

    This part is a simple wrapper around the qrcode library using the ShapelyImageFactory. If you need more
    flexibility this class might be extended or you can use the shapely output factory ShapelyImageFactory
    and qrcode directly.

    :param origin: Lower left corner of the QR code.
    :param data: Data which is to be encoded into the QR code.
    :param box_size: Size of each box.
    :param version: QR code version.
    :param error_correction: Level of error correction.
    :param border: Size of the surrounding free border.
    """
    # noinspection PyUnresolvedReferences
    from qrcode import ERROR_CORRECT_L, ERROR_CORRECT_H, ERROR_CORRECT_M, ERROR_CORRECT_Q

    def __init__(self, origin, data, box_size, version=None, error_correction=ERROR_CORRECT_M, border=0,
                 alignment='left-bottom'):
        self._qr_code = qrcode.QRCode(image_factory=ShapelyImageFactory, error_correction=error_correction,
                                      border=border, box_size=1, version=version)
        self._qr_code.add_data(data)
        self._alignment = Alignment(alignment)

        self.box_size = box_size
        self.origin = origin

    def get_shapely_object(self):
        self._qr_code.make(fit=True)
        shapely_code = self._qr_code.make_image(origin=self.origin, scale_factor=self.box_size)

        offset = self._alignment.calculate_offset(((0, 0), shapely_code.size))
        return shapely.affinity.translate(shapely_code.get_shapely_object(), *offset)


def _example():
    qr_code = QRCode([0, 0], 'A0.0', 1.0, version=1, error_correction=QRCode.ERROR_CORRECT_M)

    from gdshelpers.geometry.chip import Cell

    device = Cell('test')
    device.add_to_layer(1, qr_code)

    device.show()

    chip = Cell('optical_codes')
    chip.add_cell(device)
    chip.start_viewer()
    chip.save()


if __name__ == '__main__':
    _example()



import string
import numpy as np
import numpy.linalg as linalg


def raith_eline_dosefactor_to_datatype(dose_factor):
    """
    Convert a dose factor to a GDS datatype for Raith E-Beam writer.

    :param dose_factor: The dose factor.
    :type dose_factor: float
    :return: GDS datatype number
    :rtype: int
    """
    assert dose_factor >= 0
    return int(dose_factor * 1000)


def int_to_alphabet(num):
    """
    Convert an integer number to an alphabetic representation.

    Numbers of 0 to 25 are mapped to ``A``-``Z``, higher numbers 'count' like ``AA``, ``AB``, ..., ``AZ``, ``BA``.
    There is no upper limit on the converted number.

    :param num: Number to convert
    :return: Converted string.
    :rtype: str
    """
    assert num >= 0, 'Cannot convert negative numbers'
    numerals = string.ascii_uppercase
    b = len(numerals)
    return (num < b and numerals[num]) or int_to_alphabet((num // b) - 1) + numerals[(num % b)]


def id_to_alphanumeric(column, row):
    """
    Convert a *column*, *row* pair to an alphanumeric representation.

    :param column: Column
    :type column: int
    :param row: Row
    :type row: int
    :return: Alphanumeric representation.
    :rtype: str
    """
    return int_to_alphabet(row) + str(int(column))


def normalize_phase(phase, zero_to_two_pi=False):
    """
    Normalize a phase to be within +/- pi.

    :param phase: Phase to normalize.
    :type phase: float
    :param zero_to_two_pi: True ->  0 to 2*pi, False -> +/- pi
    :type zero_to_two_pi: bool
    :return: Normalized phase within +/- pi or 0 to 2*pi
    :rtype: float
    """

    if not zero_to_two_pi:
        return (phase + np.pi) % (2 * np.pi) - np.pi
    else:
        return (phase + 2 * np.pi) % (2 * np.pi)


def find_line_intersection(r1, angle1, r2, angle2):
    """
    Find intersection between two lines defined by point and direction.

    :param r1: Origin of the first line.
    :param angle1: Angle of the first line.
    :param r2: Origin of the second line.
    :param angle2: Angle of the second line.
    :return: Tuple of point of intersection and distances from the origins.
    :rtype: tuple
    """
    u1 = np.array([np.cos(angle1), np.sin(angle1)])
    u2 = np.array([np.cos(angle2), np.sin(angle2)])

    a = np.array([[u1[0], -u2[0]], [u1[1], -u2[1]]])
    b = r2 - r1

    # noinspection PyTypeChecker
    x = linalg.solve(a, b)
    return r1 + u1 * x[0], x
